'''Высокоуровневый API для доступа к Битрикс24'''

import re
from asyncio import Semaphore
from contextlib import asynccontextmanager, contextmanager
from typing import Iterable, Union

from . import correct_asyncio
from .srh import ServerRequestHandler
from .user_request import (BatchUserRequest, CallUserRequest,
                           GetAllUserRequest, GetByIDUserRequest)


class Bitrix:
    '''Клиент для запросов к серверу Битрикс24.'''

    '''
    Параметры:
    - `webhook: str` - URL вебхука, полученного от сервера Битрикс
    - `verbose: bool = True` - показывать ли прогрессбар при выполнении запроса
    '''

    def __init__(self, webhook: str, verbose: bool = True):
        '''
        Создает объект класса Bitrix.

        '''

        self.srh = ServerRequestHandler(webhook, verbose)

    def get_all(self, method: str, params: dict = None) -> Union[list, dict]:
        '''
        Получить полный список сущностей по запросу `method`.

        Под капотом использует параллельные запросы и автоматическое построение
        батчей, чтобы ускорить получение данных. Также самостоятельно
        обратывает постраничные ответы сервера, чтобы вернуть полный список.

        Параметры:
        - `method` - метод REST API для запроса к серверу
        - `params` - параметры для передачи методу. Используется именно тот
            формат, который указан в документации к REST API Битрикс24.
            `get_all()` не поддерживает параметры `start`, `limit` и `order`.

        Возвращает полный список сущностей, имеющихся на сервере,
        согласно заданным методу и параметрам.
        '''

        return self.srh.run(GetAllUserRequest(self.srh, method, params).run())

    def get_by_ID(self, method: str, ID_list: Iterable,
                  ID_field_name: str = 'ID', params: dict = None) -> dict:
        '''
        Получить список сущностей по запросу method и списку ID.

        Используется для случаев, когда нужны не все сущности,
        имеющиеся в базе, а конкретный список поименованных ID.
        Например, все контакты, привязанные к сделкам.

        Параметры:
        - `method` - метод REST API для запроса к серверу
        - `ID_list` - список ID
        - `ID_list_name` - название поля, которе будет подаваться в запрос для
            каждого элемента ID_list
        - `params` - параметры для передачи методу. Используется именно тот
            формат, который указан в документации к REST API Битрикс24

        Возвращает словарь вида:
        ```
        {
            ID_1: <результат запроса 1>,
            ID_2: <результат запроса 2>,
            ...
        }
        ```

        Значением элемента словаря будет результат выполнения запроса
        относительно этого ID. Это может быть, например, список связанных
        сущностей или пустой список, если не найдено ни одной привязанной
        сущности.
        '''

        return self.srh.run(GetByIDUserRequest(
            self.srh, method, params, ID_list, ID_field_name).run())

    def list_and_get(self, method_branch: str) -> dict:
        '''
        Скачать список всех ID при помощи метода *.list,
        а затем все элементы при помощи метода *.get.

        Подобный подход показывает на порядок большую скорость
        получения данных, чем `get_all()`
        с параметром `'select': ['*', 'UF_*']`.

        Параметры:
        * `method_branch: str` - группа методов к использованию, например,
        `crm.lead` или `tasks.task`

        Возвращает полное содержимое всех элементов в виде, используемом
        функцией `get_by_ID()` - словарь следующего вида:
        ```
        {
            ID_1: <словарь полей сущности с ID_1>,
            ID_2: <словарь полей сущности с ID_2>,
            ...
        }
        ```
        '''

        if not isinstance(method_branch, str):
            raise TypeError('"method_branch" should be a str')

        if re.search(r'(\.list|\.get)$', method_branch.strip().lower()):
            raise ValueError(
                '"method_branch" should not end in ".list" or ".get"')

        IDs = self.get_all(method_branch + '.list', params={'select': ['ID']})

        return self.get_by_ID(method_branch + '.get', [x['ID'] for x in IDs])

    def call(self, method: str, items: Union[dict, Iterable]):
        '''
        Вызвать метод REST API по списку элементов.

        Параметры:
        - `method` - метод REST API
        - `items` - список параметров вызываемого метода
            либо dict с параметрами для единичного вызова

        Возвращает список ответов сервера для каждого из элементов `items`
        либо просто результат для единичного вызова.
        '''

        try:
            if not (isinstance(items, dict) or
                    all(isinstance(item, dict) for item in items)):
                raise TypeError
        except TypeError:
            raise TypeError(
                'call() accepts either an iterable of params dicts or '
                'a single params dict')

        is_single_item = isinstance(items, dict)
        item_list = [items] if is_single_item else items

        request = CallUserRequest(self.srh, method, item_list)
        result = self.srh.run(request.run())

        return result[0] if is_single_item else result

    def call_batch(self, params: dict) -> dict:
        '''
        Вызвать метод `batch`.

        Параметры:
        - `params` - список параметров вызываемого метода

        Возвращает ответы сервера в формате словаря, где ключ - название
        команды, а значение - ответ сервера по этой команде.
        '''

        return self.srh.run(BatchUserRequest(self.srh, params).run())

    @contextmanager
    def slow(self, max_concurrent_requests: int = 1):
        '''Временно ограничивает количество одновременно выполняемых запросов
        к Битрикс24.'''

        if not isinstance(max_concurrent_requests, int):
            raise 'slow() argument should be only int'
        if max_concurrent_requests < 1:
            raise 'slow() argument should be >= 1'

        sem_backup = self.srh.concurrent_requests_sem
        self.srh.concurrent_requests_sem = Semaphore(max_concurrent_requests)
        yield True
        self.srh.concurrent_requests_sem = sem_backup


class BitrixAsync:
    '''
    Класс, повторяющий интерфейс класса `Bitrix`, но с асинхронными методами.

    Параметры:
    - `webhook: str` - URL вебхука, полученного от сервера Битрикс
    - `verbose: bool = True` - показывать ли прогрессбар при выполнении запроса
    '''

    def __init__(self, webhook: str, verbose: bool = True):
        '''
        Создает объект класса Bitrix.

        '''

        self.srh = ServerRequestHandler(webhook, verbose)

    async def get_all(self, method: str, params: dict = None) -> \
            Union[list, dict]:
        '''
        Получить полный список сущностей по запросу `method`.

        Под капотом использует параллельные запросы и автоматическое построение
        батчей, чтобы ускорить получение данных. Также самостоятельно
        обратывает постраничные ответы сервера, чтобы вернуть полный список.

        Параметры:
        - `method` - метод REST API для запроса к серверу
        - `params` - параметры для передачи методу. Используется
            именно тот формат, который указан в документации к REST API
            Битрикс24. `get_all()` не поддерживает параметры
            `start`, `limit` и `order`.

        Возвращает полный список сущностей, имеющихся на сервере,
        согласно заданным методу и параметрам.
        '''

        return await self.srh.run_async(
            GetAllUserRequest(self.srh, method, params).run())

    async def get_by_ID(self, method: str, ID_list: Iterable,
                        ID_field_name: str = 'ID',
                        params: dict = None) -> list:
        '''
        Получить список сущностей по запросу `method` и списку ID.

        Используется для случаев, когда нужны не все сущности,
        имеющиеся в базе, а конкретный список поименованных ID.
        Например, все контакты, привязанные к сделкам.

        Параметры:
        - `method` - метод REST API для запроса к серверу
        - `ID_list` - список ID
        - `ID_list_name` - название поля, которе будет подаваться в запрос для
            каждого элемента ID_list
        - `params` - параметры для передачи методу. Используется именно тот
            формат, который указан в документации к REST API Битрикс24

        Возвращает словарь вида:
        ```
        {
            ID_1: <результат запроса 1>,
            ID_2: <результат запроса 2>,
            ...
        }
        ```

        Значением элемента словаря будет результат выполнения запроса
        относительно этого ID. Это может быть, например, список связанных
        сущностей или пустой список, если не найдено ни одной привязанной
        сущности.
        '''

        return await self.srh.run_async(GetByIDUserRequest(
            self.srh, method, params, ID_list, ID_field_name).run())

    async def list_and_get(self, method_branch: str) -> dict:
        '''
        Скачать список всех ID при помощи метода *.list,
        а затем все элементы при помощи метода *.get.

        Подобный подход показывает на порядок большую скорость
        получения данных, чем `get_all()`
        с параметром `'select': ['*', 'UF_*']`.

        Параметры:
        * `method_branch: str` - группа методов к использованию, например,
        `crm.lead` или `tasks.task`

        Возвращает полное содержимое всех элементов в виде, используемом
        функцией `get_by_ID()` - словарь следующего вида:
        ```
        {
            ID_1: <словарь полей сущности с ID_1>,
            ID_2: <словарь полей сущности с ID_2>,
            ...
        }
        ```
        '''

        if not isinstance(method_branch, str):
            raise TypeError('"method_branch" should be a str')

        if re.search(r'(\.list|\.get)$', method_branch.strip().lower()):
            raise ValueError(
                '"method_branch" should not end in ".list" or ".get"')

        IDs = await self.get_all(
            method_branch + '.list', params={'select': ['ID']})

        return await self.get_by_ID(
            method_branch + '.get', [x['ID'] for x in IDs])

    async def call(self, method: str, items: Union[dict, Iterable]):
        '''
        Вызвать метод REST API по списку элементов.

        Параметры:
        - `method` - метод REST API
        - `items` - список параметров вызываемого метода
            либо dict с параметрами для единичного вызова

        Возвращает список ответов сервера для каждого из элементов `items`
        либо просто результат для единичного вызова.
        '''

        try:
            if not (isinstance(items, dict) or
                    all(isinstance(item, dict) for item in items)):
                raise TypeError
        except TypeError:
            raise TypeError(
                'call() accepts either an iterable of params dicts or '
                'a single params dict')

        is_single_item = isinstance(items, dict)
        item_list = [items] if is_single_item else items

        request = CallUserRequest(self.srh, method, item_list)
        result = await self.srh.run_async(request.run())

        return result[0] if is_single_item else result

    async def call_batch(self, params: dict) -> dict:
        '''
        Вызвать метод `batch`.

        Параметры:
        - `params` - список параметров вызываемого метода

        Возвращает ответы сервера в формате словаря, где ключ - название
        команды, а значение - ответ сервера по этой команде.
        '''

        return await self.srh.run_async(
            BatchUserRequest(self.srh, params).run())

    @asynccontextmanager
    async def slow(self, max_concurrent_requests: int = 1):
        '''Временно ограничивает количество одновременно выполняемых запросов
        к Битрикс24.'''

        if not isinstance(max_concurrent_requests, int):
            raise 'slow() argument should be only int'
        if max_concurrent_requests < 1:
            raise 'slow() argument should be >= 1'

        sem_backup = self.srh.concurrent_requests_sem
        self.srh.concurrent_requests_sem = Semaphore(max_concurrent_requests)
        yield True
        self.srh.concurrent_requests_sem = sem_backup



from asyncio import ensure_future, wait
from itertools import chain

from more_itertools import chunked

from .server_response import ServerResponse
from .srh import BITRIX_MAX_BATCH_SIZE, ServerRequestHandler
from .utils import http_build_query


class MultipleServerRequestHandler:

    def __init__(self, srh: ServerRequestHandler, method, item_list,
                 real_len=None, real_start=0):
        self.srh = srh
        self.method = method
        self.item_list = item_list
        self.real_len = real_len or len(item_list)
        self.real_start = real_start
        self.results = []

    async def run(self):
        self.pbar = self.srh.get_pbar(self.real_len, self.real_start)

        self.task_iterator = self.generate_a_task()
        self.tasks = set()
        self.top_up_tasks()

        while self.tasks:
            done, _ = await wait(self.tasks)

            for done_task in done:
                batch_response = done_task.result()
                unwrapped_result = ServerResponse(batch_response.result).result
                extracted_len = self.extract_result_from_batch_response(
                    unwrapped_result)
                self.pbar.update(extracted_len)

            self.tasks -= done
            self.top_up_tasks()

        self.pbar.close()
        return self.results

    def top_up_tasks(self):
        '''Добавляем в self.tasks столько задач, сколько свободных слотов для
        запросов есть сейчас в self.srh.'''

        while len(self.tasks) < self.srh.concurrent_requests_sem._value:
            try:
                self.tasks.add(next(self.task_iterator))
            except StopIteration:
                break

    def generate_a_task(self):
        '''Объединяем элементы item_list в батчи и по одной создаем и
        возвращаем задачи asyncio с запросами к серверу для каждого батча.'''

        batches = [{
            'halt': 0,
            'cmd': {
                self.batch_command_label(i, item):
                f'{self.method}?{http_build_query(item)}'
                for i, item in enumerate(next_batch)
            }}
            for next_batch in chunked(self.item_list, BITRIX_MAX_BATCH_SIZE)]

        for batch in batches:
            yield ensure_future(self.srh.single_request('batch', batch))

    def batch_command_label(self, i, item):
        return f'cmd{i}'

    def extract_result_from_batch_response(self, unwrapped_result):
        '''Добавляет `unwrapped_result` в `self.results` и возвращает
        длину добавленного списка результатов'''
        result_list = list(unwrapped_result.values())
        if type(result_list[0]) == list:
            result_list = list(chain(*result_list))
        self.results.extend(result_list)
        return len(result_list)


class MultipleServerRequestHandlerPreserveIDs(MultipleServerRequestHandler):

    def __init__(self, srh, method, item_list, ID_field):
        super().__init__(srh, method, item_list)
        self.ID_field = ID_field
        self.results = {}

    def batch_command_label(self, i, item):
        return item[self.ID_field]

    def extract_result_from_batch_response(self, unwrapped_result):
        self.results.update(unwrapped_result)
        return len(unwrapped_result)



import asyncio
import time
from collections import deque
from contextlib import asynccontextmanager

import aiohttp
from aiohttp.client_exceptions import (ClientPayloadError, ClientResponseError,
                                       ServerDisconnectedError)
from tqdm import tqdm

from .server_response import ServerResponse
from .utils import _url_valid, retry


BITRIX_POOL_SIZE = 50
BITRIX_RPS = 2.0
BITRIX_MAX_BATCH_SIZE = 50
BITRIX_MAX_CONCURRENT_REQUESTS = 20


class ServerError(Exception):
    pass


class ServerRequestHandler():
    '''
    Используется для контроля скорости доступа к серверам Битрикс.

    Основная цель - вести учет количества запросов, которые можно передать
    серверу Битрикс без получения ошибки `5XX`.

    Используется как контекстный менеджер, оборачивающий несколько
    последовательных запросов к серверу.
    '''

    def __init__(self, webhook, verbose):
        self.webhook = self._standardize_webhook(webhook)
        self._verbose = verbose

        self.requests_per_second = BITRIX_RPS
        self.pool_size = BITRIX_POOL_SIZE

        self.active_runs = 0
        self.session = None

        # rr - requests register - список отправленных запросов к серверу
        self.rr = deque()

        self.concurrent_requests_sem = asyncio.Semaphore(
            BITRIX_MAX_CONCURRENT_REQUESTS)

    def _standardize_webhook(self, webhook):
        '''Приводит `webhook` к стандартному виду.'''

        if not isinstance(webhook, str):
            raise TypeError(f'Webhook should be a {str}')

        webhook = webhook.lower().strip()

        if not _url_valid(webhook):
            raise ValueError('Webhook is not a valid URL')

        if webhook[-1] != '/':
            webhook += '/'

        return webhook

    def run(self, coroutine):
        '''Запускает `coroutine`, оборачивая его в `run_async()`.'''

        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            asyncio.set_event_loop(asyncio.new_event_loop())
            loop = asyncio.get_event_loop()

        return loop.run_until_complete(self.run_async(coroutine))

    async def run_async(self, coroutine):
        '''Запускает `coroutine`, создавая и прекращая сессию
        при необходимости.'''

        async with self.handle_sessions():
            return await coroutine

    @asynccontextmanager
    async def handle_sessions(self):
        '''Открывает и закрывает сессию в зависимости от наличия
        активных запросов.'''

        if not self.active_runs and (not self.session or self.session.closed):
            self.session = aiohttp.ClientSession(raise_for_status=True)
        self.active_runs += 1

        yield True

        self.active_runs -= 1
        if not self.active_runs and self.session and not self.session.closed:
            await self.session.close()

    @retry(exceptions=[ClientPayloadError, ServerDisconnectedError,
                       ServerError])
    async def single_request(self, method, params=None):
        '''Делает единичный запрос к серверу, ожидая при необходимости.'''

        try:
            async with self.concurrent_requests_sem, self.acquire(), \
                    self.session.post(url=self.webhook + method, json=params) \
                    as response:
                return ServerResponse(await response.json(encoding='utf-8'))
        except ClientResponseError as error:
            if error.status // 100 == 5:  # ошибки вида 5XX
                raise ServerError('The server returned an error') from error
            else:
                raise

    @asynccontextmanager
    async def acquire(self):
        '''Ожидает, пока не станет безопасно делать запрос к серверу.'''

        # если пул заполнен, ждать
        if len(self.rr) >= self.pool_size:
            time_from_last_request = time.monotonic() - self.rr[0]
            time_to_wait = 1 / self.requests_per_second - \
                time_from_last_request
            if time_to_wait > 0:
                await asyncio.sleep(time_to_wait)

        # зарегистрировать запрос в очереди
        cur_time = time.monotonic()
        self.rr.appendleft(cur_time)

        # отдать управление
        yield

        # подчистить пул
        trim_time = cur_time - self.pool_size / self.requests_per_second
        while self.rr and self.rr[len(self.rr) - 1] < trim_time:
            self.rr.pop()

    def get_pbar(self, real_len, real_start):
        '''Возвращает прогресс бар `tqdm()` или пустышку,
        если `self.verbose is False`.'''

        class MutePBar():

            def update(self, i):
                pass

            def close(self):
                pass

        if self._verbose:
            return tqdm(total=real_len, initial=real_start)
        else:
            return MutePBar()



import pickle
from typing import Iterable
import warnings
from collections import ChainMap
from collections.abc import Sequence

from .mult_request import (MultipleServerRequestHandler,
                           MultipleServerRequestHandlerPreserveIDs)
from .server_response import ServerResponse
from .srh import ServerRequestHandler

BITRIX_PAGE_SIZE = 50


class UserRequestAbstract():

    def __init__(self, srh: ServerRequestHandler, method: str,
                 params: dict = None):
        self.srh = srh
        self.method = method
        self.st_method = self.standardized_method(method)
        self.params = params
        self.st_params = self.standardized_params(params)
        self.check_special_limitations()

    def standardized_method(self, method):
        if not method:
            raise TypeError('Method cannot be empty')

        if not isinstance(method, str):
            raise TypeError('Method should be a str')

        method = method.lower().strip()

        if method == 'batch':
            raise ValueError(
                "Method cannot be 'batch'. Use call_batch() instead.")

        return method

    def standardized_params(self, p):
        if p is None:
            return None

        if not isinstance(p, dict):
            raise TypeError('Params agrument should be a dict')

        for key, __ in p.items():
            if not isinstance(key, str):
                raise TypeError('Keys in params argument should be strs')

        p = {key.upper().strip(): value for key, value in p.items()}

        self.check_expected_clause_types(p)

        return p

    def check_expected_clause_types(self, p):
        EXPECTED_TYPES = {
            'SELECT': list,
            'HELT': int,
            'CMD': dict,
            'LIMIT': int,
            'ORDER': dict,
            'FILTER': dict,
            'START': int,
            'FIELDS': dict
        }

        # check for allowed types of key values
        for clause_key, clause_value in p.items():
            if clause_key in EXPECTED_TYPES:
                expected_type = EXPECTED_TYPES[clause_key]

                type_ok = isinstance(clause_value, expected_type)
                if expected_type == list:
                    list_error = not any(
                        isinstance(clause_value, x) for x in [list, tuple, set]
                    )
                else:
                    list_error = False

                if not type_ok or list_error:
                    raise TypeError(
                        f'Clause "{clause_key}" should be '
                        f'of type {expected_type}, '
                        f'but its type is {type(clause_value)}')

    def check_special_limitations(self):
        raise NotImplementedError


class GetAllUserRequest(UserRequestAbstract):

    def check_special_limitations(self):
        if self.st_params and not set(self.st_params.keys()).isdisjoint(
            {'START', 'LIMIT', 'ORDER'}
        ):
            raise ValueError("get_all() doesn't support parameters "
                             "'start', 'limit' or 'order'")

    async def run(self):
        self.add_order_parameter()

        await self.make_first_request()

        if self.first_response.more_results_expected():
            await self.make_remaining_requests()
            self.dedup_results()

        return self.results

    def add_order_parameter(self):
        # необходимо установить порядок сортировки, иначе сортировка
        # будет рандомная и сущности будут повторяться на разных страницах

        order_clause = {'order': {'ID': 'ASC'}}

        if self.params:
            if 'ORDER' not in self.st_params:
                self.params.update(order_clause)
        else:
            self.params = order_clause

    async def make_first_request(self):
        self.first_response = await self.srh.single_request(
            self.method, self.params)
        self.results, self.total = \
            self.first_response.result, self.first_response.total

    async def make_remaining_requests(self):
        item_list = (
            ChainMap({'start': start}, self.params)
            for start in range(len(self.results), self.total, BITRIX_PAGE_SIZE)
        )
        remaining_results = await MultipleServerRequestHandler(
                self.srh,
                method=self.method,
                item_list=item_list,
                real_len=self.total,
                real_start=len(self.results)
            ).run()

        self.results.extend(remaining_results)

    def dedup_results(self):
        # дедупликация через сериализацию, превращение в set и десериализацию
        self.results = (
            [pickle.loads(y) for y in {pickle.dumps(x) for x in self.results}]
            if self.results
            else []
        )

        if len(self.results) != self.total:
            warnings.warn(
                f"Number of results returned ({len(self.results)}) "
                f"doesn't equal 'total' from the server reply ({self.total})",
                RuntimeWarning)


class GetByIDUserRequest(UserRequestAbstract):

    def __init__(self, srh: ServerRequestHandler, method: str, params: dict,
                 ID_list: Iterable, ID_field_name: str):
        self.ID_list = ID_list
        self.ID_field_name = ID_field_name.strip()
        super().__init__(srh, method, params)

    def check_special_limitations(self):
        if self.st_params and 'ID' in self.st_params.keys():
            raise ValueError("get_by_ID() doesn't support parameter 'ID' "
                             "within the 'params' argument")

        try:
            iter(self.ID_list)
        except TypeError:
            raise TypeError("get_by_ID(): 'ID_list' should be iterable")

        for ID in self.ID_list:
            if not isinstance(ID, (str, int)):
                raise TypeError(
                    "get_by_ID(): 'ID_list' should contain only ints or strs")

    async def run(self):
        if self.list_empty():
            return []

        self.prepare_item_list()

        results = await MultipleServerRequestHandlerPreserveIDs(
            self.srh,
            self.method,
            self.item_list,
            ID_field=self.ID_field_name
        ).run()

        return results

    def list_empty(self):
        return len(self.ID_list) == 0

    def prepare_item_list(self):
        if self.params:
            self.item_list = [
                ChainMap({self.ID_field_name: ID}, self.params)
                for ID in self.ID_list
            ]
        else:
            self.item_list = [
                {self.ID_field_name: ID}
                for ID in self.ID_list
            ]


class CallUserRequest(GetByIDUserRequest):

    def __init__(self, srh: ServerRequestHandler, method: str,
                 item_list: Iterable):
        self.item_list = item_list
        super().__init__(srh, method, None, None, '__order')

    def check_special_limitations(self):
        try:
            iter(self.item_list)
        except TypeError:
            raise TypeError("call(): 'item_list' should be iterable")

    async def run(self):
        results = await super().run()

        # убираем поле с порядковым номером из результатов
        return tuple() if self.list_empty() else tuple(results.values())

    def list_empty(self):
        return len(self.item_list) == 0

    def prepare_item_list(self):
        # добавим порядковый номер
        self.item_list = [
            ChainMap(item, {self.ID_field_name: 'order' + str(i)})
            for i, item in enumerate(self.item_list)
        ]


class BatchUserRequest(UserRequestAbstract):

    def __init__(self, srh, params):
        super().__init__(srh, 'batch', params)

    def standardized_method(self, method):
        return 'batch'

    def check_special_limitations(self):
        if not self.st_params:
            raise ValueError("Params for a batch call can't be empty")

        if {'HALT', 'CMD'} != self.st_params.keys():
            raise ValueError(
                "Params for a batch call should contain only 'halt' and 'cmd' "
                "clauses at the highest level")

        if not isinstance(self.st_params['CMD'], dict):
            raise ValueError("'cmd' clause should contain a dict")

    async def run(self):
        response = await self.srh.single_request(self.method, self.params)
        return ServerResponse(response.result).result



class ServerResponse():

    def __init__(self, response):
        self.response = response
        self.check_for_errors()
        
        
    def check_for_errors(self):
        if self.result_error:
            raise RuntimeError(f'The server reply contained an error: {self.result_error}')
        
        
    def __getattr__(self, item):
        if item in self.response.keys():
            return self.response[item]
        else:
            return None
        
        
    def more_results_expected(self):
        return self.total and self.total > 50 and self.total != len(self.result)


from functools import wraps
from time import sleep
from urllib.parse import quote, urlparse


def _url_valid(url):
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc, result.path])
    except:
        return False


def http_build_query(params, convention="%s"):
    if len(params) == 0:
        return ""

    output = ""
    for key in params.keys():

        if type(params[key]) is dict:

            output += http_build_query(params[key], convention % key + "[%s]")

        elif type(params[key]) is list:

            new_params = {str(i): element for i, element
                          in enumerate(params[key])}

            output += http_build_query(
                new_params, convention % key + "[%s]")

        else:

            val = quote(str(params[key]))
            key = quote(key)
            output = output + convention % key + "=" + val + "&"

    return output


def retry(exceptions, total_tries=3, initial_wait=0.5, backoff_factor=2):
    """
    Calling the decorated function applying an exponential backoff.
    Args:
        exceptions: Exeption(s) that trigger a retry, can be a tuple
        total_tries: Total tries
        initial_wait: Time to first retry
        backoff_factor: Backoff multiplier (e.g. value of 2 will double
            the delay each retry).
    """
    def retry_decorator(f):

        @wraps(f)
        def func_with_retries(*args, **kwargs):

            _tries, _delay = total_tries + 1, initial_wait

            while _tries > 1:
                try:
                    return f(*args, **kwargs)
                except tuple(exceptions):
                    _tries -= 1
                    if _tries == 1:
                        raise
                    sleep(_delay)
                    _delay *= backoff_factor

        return func_with_retries

    return retry_decorator
